"""The module implementing the Client class."""

import typing as T

import dateutil
import qibo
import tabulate
from packaging.version import Version

from . import constants
from .config_logging import logger
from .exceptions import JobPostServerError
from .qibo_job import QiboJob
from .utils import QiboApiRequest


class Client:
    """Class to manage the interaction with the remote server."""

    def __init__(self, token: str, url: str = constants.BASE_URL):
        """
        :param token: the authentication token associated to the webapp user
        :type token: str
        :param url: the server address
        :type url: str
        """
        self.token = token
        self.headers = {"x-api-token": token}
        self.base_url = url

        self.pid = None
        self.results_folder = None
        self.results_path = None

    def check_client_server_qibo_versions(self):
        """Check that client and server qibo package installed versions match.

        Raise assertion error if the two versions are not the same.
        """
        url = self.base_url + "/api/qibo_version/"
        response = QiboApiRequest.get(
            url,
            timeout=constants.TIMEOUT,
            keys_to_check=["server_qibo_version", "minimum_client_qibo_version"],
        )

        qibo_server_version = Version(response.json()["server_qibo_version"])
        qibo_minimum_client_version = Version(
            response.json()["minimum_client_qibo_version"]
        )

        qibo_client_version = Version(qibo.__version__)
        msg = (
            "The qibo-client package requires an installed qibo package version"
            f">={qibo_minimum_client_version}, the local qibo "
            f"version is {qibo_client_version}"
        )
        assert qibo_client_version >= qibo_minimum_client_version, msg

        if qibo_client_version < qibo_server_version:
            logger.warning(
                "Local Qibo package version does not match the server one, please "
                "upgrade: %s -> %s",
                qibo_client_version,
                qibo_server_version,
            )

    def run_circuit(
        self,
        circuit: qibo.Circuit,
        device: str,
        project: str = "personal",
        nshots: T.Optional[int] = None,
        verbatim: bool = False,
    ) -> T.Optional[
        T.Union[
            qibo.result.QuantumState,
            qibo.result.MeasurementOutcomes,
            qibo.result.CircuitResult,
        ]
    ]:
        """Run circuit on the cluster.

        :param circuit: the QASM representation of the circuit to run
        :type circuit: Circuit
        :param device: the device to run the circuit on.
        :type device: str
        :type project: the project to run the circuit on.
        :type project: str
        :param nshots: number of shots, mandatory for non-simulation devices, defaults to `nshots=100` for simulation partitions
        :type nshots: int
        :param verbatim: If True, attempts to run the circuit without any transpilation. Defaults to False.
        :type verbatim: bool
        :param wait_for_results: whether to let the client hang until server results are ready or not. Defaults to True.
        :type wait_for_results: bool

        :return:
            the result of the computation. None if the job
            raised an error.
        :rtype: Optional[QiboJobResult]
        """
        self.check_client_server_qibo_versions()
        logger.info("Post new circuit on the server")
        job = self._post_circuit(circuit, device, project, nshots, verbatim)

        logger.info("Job posted on server with pid %s", self.pid)
        return job

    def _post_circuit(
        self,
        circuit: qibo.Circuit,
        device: str,
        project: str,
        nshots: T.Optional[int] = None,
        verbatim: bool = False,
    ) -> QiboJob:
        url = self.base_url + "/api/jobs/"

        payload = {
            "circuit": circuit.raw,
            "nshots": nshots,
            "device": device,
            "project": project,
            "verbatim": verbatim,
        }
        response = QiboApiRequest.post(
            url,
            headers=self.headers,
            json=payload,
            timeout=constants.TIMEOUT,
        )
        result = response.json()

        self.pid = result.get("pid")

        if self.pid is None:
            raise JobPostServerError(result["detail"])

        return QiboJob(
            base_url=self.base_url,
            headers=self.headers,
            pid=self.pid,
            circuit=circuit.raw,
            nshots=nshots,
            device=device,
        )

    def print_quota_info(self):
        """Logs the formatted user quota info table."""
        url = self.base_url + "/api/disk_quota/"

        response = QiboApiRequest.get(
            url,
            headers=self.headers,
            timeout=constants.TIMEOUT,
        )

        disk_quota = response.json()[0]

        url = self.base_url + "/api/projectquotas/"

        response = QiboApiRequest.get(
            url,
            headers=self.headers,
            timeout=constants.TIMEOUT,
        )

        projectquotas = response.json()

        message = (
            f"User: {disk_quota['user']['email']}\n"
            f"Disk quota left [KBs]: {disk_quota['kbs_left']:.2f} / {disk_quota['kbs_max']:.2f}\n"
        )

        rows = [
            (
                t["project"]["name"],
                t["partition"]["name"],
                t["partition"]["max_num_qubits"],
                t["partition"]["hardware_type"],
                t["partition"]["description"],
                t["partition"]["status"],
                t["seconds_left"],
                t["shots_left"],
                t["jobs_left"],
            )
            for t in projectquotas
        ]
        message += tabulate.tabulate(
            rows,
            headers=[
                "Project Name",
                "Device Name",
                "Qubits",
                "Type",
                "Description",
                "Status",
                "Time Left [s]",
                "Shots Left",
                "Jobs Left",
            ],
        )
        logger.info(message)

    def print_job_info(self):
        """Logs the formatted user quota info table."""
        url = self.base_url + "/api/jobs/"

        response = QiboApiRequest.get(
            url,
            headers=self.headers,
            timeout=constants.TIMEOUT,
        )

        def format_date(dt: str) -> str:
            dt = dateutil.parser.isoparse(dt)
            return dt.strftime("%Y-%m-%d %H:%M:%S")

        jobs = response.json()
        if not len(jobs):
            logger.info("No jobs found in database for user")
            return None

        user_set = {job["user"]["email"] for job in jobs}
        if len(user_set) > 1:
            raise ValueError(
                "The `/api/jobs/` endpoint returned info about " "multiple accounts."
            )
        user = list(user_set)[0]

        rows = [
            (
                job["pid"],
                format_date(job["created_at"]),
                format_date(job["updated_at"]),
                job["status"],
                job["result_path"],
            )
            for job in response.json()
        ]
        message = f"User: {user}\n" + tabulate.tabulate(
            rows, headers=["Pid", "Created At", "Updated At", "Status", "Results"]
        )
        logger.info(message)

    def get_job(self, pid: str) -> QiboJob:
        """Retrieves the job from the unique process id.

        :param pid: the job's process identifier
        :type pid: str

        :return: the requested QiboJob object
        :rtype: QiboJob
        """
        job = QiboJob(base_url=self.base_url, pid=pid)
        job.refresh()
        return job

    def delete_job(self, pid: str):
        """Removes the given job from the web server.

        :param pid: the job's process identifier
        :type pid: str
        """
        job = QiboJob(base_url=self.base_url, pid=pid)
        return job.delete()



import tarfile
import tempfile
import time
import typing as T
from enum import Enum
from pathlib import Path

import qibo
import requests

from . import constants
from .config_logging import logger
from .utils import QiboApiRequest


def convert_str_to_job_status(status: str):
    return next((s for s in QiboJobStatus if s.value == status), None)


class QiboJobStatus(Enum):
    QUEUEING = "queueing"
    PENDING = "pending"
    RUNNING = "running"
    POSTPROCESSING = "postprocessing"
    SUCCESS = "success"
    ERROR = "error"


def _write_stream_to_tmp_file(stream: T.Iterable) -> Path:
    """Write chunk of bytes to temporary file.

    The tmp_path should be closed manually.

    :param stream: the stream of bytes chunks to be saved on disk
    :type stream: Iterable

    :return: the name of the tempo

    """
    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
        for chunk in stream:
            if chunk:
                tmp_file.write(chunk)
        archive_path = tmp_file.name
    return Path(archive_path)


def _extract_archive_to_folder(source_archive: Path, destination_folder: Path):
    with tarfile.open(source_archive, "r:gz") as archive:
        archive.extractall(destination_folder)


def _save_and_unpack_stream_response_to_folder(
    stream: T.Iterable, results_folder: Path
):
    """Save the stream to a given folder.

    Internally, save the stream to a temporary archive and extract its contents
    to the target folder.

    :param stream: the iterator containing the response content
    :type stream: Iterable
    :param results_folder: the local path to the results folder
    :type results_folder: Path
    """
    archive_path = _write_stream_to_tmp_file(stream)

    _extract_archive_to_folder(archive_path, results_folder)

    # clean up temporary file
    archive_path.unlink()


class QiboJob:
    def __init__(
        self,
        pid: str,
        base_url: str = constants.BASE_URL,
        headers: T.Dict[str, str] = None,
        circuit: T.Optional[qibo.Circuit] = None,
        nshots: T.Optional[int] = None,
        device: T.Optional[str] = None,
    ):
        self.base_url = base_url
        self.headers = headers
        self.pid = pid
        self.circuit = circuit
        self.nshots = nshots
        self.device = device

        self._status = None

    def refresh(self):
        """Refreshes job information from server.

        This method does not query the results from server.
        """
        url = self.base_url + f"/api/jobs/{self.pid}/"
        response = QiboApiRequest.get(
            url,
            headers=self.headers,
            timeout=constants.TIMEOUT,
            keys_to_check=["circuit", "nshots", "projectquota", "status"],
        )

        info = response.json()
        if info is not None:
            self._update_job_info(info)

    def _update_job_info(self, info: T.Dict):
        self.circuit = info.get("circuit")
        self.nshots = info.get("nshots")
        self.device = info["projectquota"]["partition"]["name"]
        self._status = convert_str_to_job_status(info["status"])

    def status(self) -> QiboJobStatus:
        url = self.base_url + f"/api/jobs/{self.pid}/"
        response = QiboApiRequest.get(
            url,
            headers=self.headers,
            timeout=constants.TIMEOUT,
            keys_to_check=["status"],
        )
        status = response.json()["status"]
        self._status = convert_str_to_job_status(status)
        return self._status

    def running(self) -> bool:
        if self._status is None:
            self.refresh()
        return self._status is QiboJobStatus.RUNNING

    def success(self) -> bool:
        if self._status is None:
            self.refresh()
        return self._status is QiboJobStatus.SUCCESS

    def result(
        self, wait: int = 5, verbose: bool = False
    ) -> T.Optional[qibo.result.QuantumState]:
        """Send requests to server checking whether the job is completed.

        This function populates the `Client.results_folder` and
        `Client.results_path` attributes.

        :return: the numpy array with the results of the computation.
                 None if the job raised an error.
        :rtype: T.Optional[np.ndarray]
        """
        # @TODO: here we can use custom logger levels instead of if statement
        response, job_status = self._wait_for_response_to_get_request(wait, verbose)

        # create the job results folder
        self.results_folder = constants.RESULTS_BASE_FOLDER / self.pid
        self.results_folder.mkdir(parents=True, exist_ok=True)

        # Save the stream to disk
        try:
            _save_and_unpack_stream_response_to_folder(
                response.iter_content(), self.results_folder
            )
        except tarfile.ReadError as err:
            logger.error("Catched tarfile ReadError: %s", err)
            logger.error(
                "The received file is not a valid gzip "
                "archive, the result might have to be inspected manually. Find "
                "the file at `%s`",
                self.results_folder.as_posix(),
            )
            return None

        if job_status == QiboJobStatus.ERROR:
            out_log_path = self.results_folder / "stdout.log"
            stdout = out_log_path.read_text() if out_log_path.is_file() else "-"

            err_log_path = self.results_folder / "stderr.log"
            stderr = err_log_path.read_text() if err_log_path.is_file() else "-"

            logger.error(
                "Job exited with error\n\nStdout:\n%s\n\nStderr:\n%s",
                stdout,
                stderr,
            )

            return None

        self.results_path = self.results_folder / "results.npy"
        return qibo.result.load_result(self.results_path)

    def _wait_for_response_to_get_request(
        self, seconds_between_checks: T.Optional[int] = None, verbose: bool = False
    ) -> T.Tuple[requests.Response, QiboJobStatus]:
        """Wait until the server completes the computation and return the response.

        :param url: the endpoint to make the request
        :type url: str

        :return: the response of the get request
        :rtype: requests.Response
        :return: the completed job response status
        :rtype: QiboJobStatus
        """
        if seconds_between_checks is None:
            seconds_between_checks = constants.SECONDS_BETWEEN_CHECKS

        is_job_finished = self.status() not in [
            QiboJobStatus.SUCCESS,
            QiboJobStatus.ERROR,
        ]
        if not verbose and is_job_finished:
            logger.info("Please wait until your job is completed...")

        url = self.base_url + f"/api/jobs/result/{self.pid}/"

        while True:
            response = QiboApiRequest.get(
                url, headers=self.headers, timeout=constants.TIMEOUT
            )
            job_status = convert_str_to_job_status(response.headers["Job-Status"])

            if verbose and job_status == QiboJobStatus.QUEUEING:
                logger.info("Job QUEUEING")
            if verbose and job_status == QiboJobStatus.PENDING:
                logger.info("Job PENDING")
            if verbose and job_status == QiboJobStatus.RUNNING:
                logger.info("Job RUNNING")
            if verbose and job_status == QiboJobStatus.POSTPROCESSING:
                logger.info("Job POSTPROCESSING")
            if job_status in [QiboJobStatus.SUCCESS, QiboJobStatus.ERROR]:
                if verbose:
                    logger.info("Job COMPLETED")
                return response, job_status
            time.sleep(seconds_between_checks)

    def delete(self) -> str:
        url = self.base_url + f"/api/jobs/{self.pid}/"
        response = QiboApiRequest.delete(
            url, headers=self.headers, timeout=constants.TIMEOUT
        )
        return response.json()["detail"]



import typing as T

import requests

from .exceptions import JobApiError, MalformedResponseError


def check_json_response_has_keys(response_json: T.Dict, keys: T.List[str]):
    """Check that the response body contains certain keys.

    :param response_json: the server json response
    :type response_json: Dict
    :param keys: the keys to be checked in the response body
    :type keys: List[str]

    :raises MalformedResponseError:
        if the server response does not contain all the expected keys.
    """
    response_keys = set(response_json.keys())
    expected_keys = set(keys)
    missing_keys = expected_keys.difference(response_keys)

    if len(missing_keys):
        raise MalformedResponseError(
            f"The server response is missing the following keys: {' '.join(missing_keys)}"
        )


def _request_and_status_check(request_fn, *args, **kwargs):
    try:
        response = request_fn(*args, **kwargs)
        response.raise_for_status()
    except requests.HTTPError:
        raise JobApiError(response.status_code, response.json().get("detail"))

    return response


def _make_request(request_fn, keys_to_check, *args, **kwargs) -> requests.Response:
    response = _request_and_status_check(request_fn, *args, **kwargs)
    if keys_to_check is not None:
        check_json_response_has_keys(response.json(), keys_to_check)
    return response


class QiboApiRequest:

    @staticmethod
    def get(
        endpoint: str,
        params: T.Optional[T.Dict] = None,
        headers: T.Optional[T.Dict] = None,
        timeout: T.Optional[float] = None,
        keys_to_check: T.Optional[T.List[str]] = None,
    ) -> requests.Response:
        return _make_request(
            requests.get,
            keys_to_check,
            endpoint,
            params=params,
            headers=headers,
            timeout=timeout,
        )

    @staticmethod
    def post(
        endpoint: str,
        headers: T.Optional[T.Dict] = None,
        json: T.Optional[T.Dict] = None,
        timeout: T.Optional[float] = None,
        keys_to_check: T.Optional[T.List[str]] = None,
    ) -> requests.Response:
        return _make_request(
            requests.post,
            keys_to_check,
            endpoint,
            headers=headers,
            json=json,
            timeout=timeout,
        )

    @staticmethod
    def delete(
        endpoint: str,
        timeout: T.Optional[float] = None,
        headers: T.Optional[T.Dict] = None,
        keys_to_check: T.Optional[T.List[str]] = None,
    ) -> requests.Response:
        return _make_request(
            requests.delete, keys_to_check, endpoint, headers=headers, timeout=timeout
        )



import logging
import os

# configure logger
logging.basicConfig(format="[%(asctime)s] %(levelname)s: %(message)s")
logger = logging.getLogger(__name__)
logging_level = os.environ.get("QIBO_CLIENT_LOGGER_LEVEL", logging.INFO)
logger.setLevel(logging_level)



import os
from pathlib import Path

RESULTS_BASE_FOLDER = Path(os.environ.get("RESULTS_BASE_FOLDER", "/tmp/qibo_client"))
SECONDS_BETWEEN_CHECKS = os.environ.get("SECONDS_BETWEEN_CHECKS", 2)

BASE_URL = "https://cloud.qibo.science"
TIMEOUT = 60
